BeginPackage["FormRun`"]

$FormCalc = {1, 0}

$FormCalcVersionNumber = 1.0

$FormRunVersion = "FormRun 1.0 (6 Jun 2019)"

FormRun::usage =
"FormRun[expr, (decl), (cmd)] sends expr to FORM for expansion,
optionally applying extra FORM declarations decl and commands cmd."

FilePrefix::usage =
"FilePrefix is an option of FormRun.  It specifies the first part of
the name of the FORM file, i.e. the 'fr-' in fr-1.frm."

EditCode::usage =
"EditCode is a debugging option of FormRun.
It determines editing of the intermediate FORM code.
True invokes the $Editor command, which is supposed to detach from
FormCalc, i.e. the FORM process continues with the unedited code.
Modal invokes the $EditorModal command, which is supposed to be
modal (non-detached), i.e. continues only after the editor is closed,
thus continuing with possibly modified FORM code."

RetainFile::usage =
"RetainFile is a debugging option of FormRun.
When set to True, the temporary file containing the FORM code is 
not removed after running FORM."

$FormCmd::usage =
"$FormCmd specifies the invocation of the FORM executable. 
Arguments are separated by a vertical bar (|)."

$FormRunDir = DirectoryName[
  $InputFileName /. HoldPattern[$InputFileName] :>
    (File /. FileInformation[System`Private`FindFile[$Input]]) ]

$FormRunBin = ToFileName[{$FormRunDir, $SystemID}]

{FormGet, root$}

Begin["`Private`"]

Print[$FormRunVersion]

Install[ToFileName[$FormRunBin, "FormGet"]]

$FormCmd = ToFileName[$FormRunBin, "tform"] <>
  "|-w" <> ToString[Max[1, $ProcessorCount]]

FormSetup = "\
#-\n\
#:SmallSize 5000000\n\
#:LargeSize 20000000\n\
#:WorkSpace 50000000\n\
#:MaxTermSize 300000\n\
#:TermsInSmall 30000\n\
format 75;\n\
off stats;\n\n\
auto s a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z;\n\
auto s A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z;\n\n";

If[ StringMatchQ[$SystemID, "Windows*"],
  $EditorModal = "notepad ``";
  $Editor := "start /b " <> $EditorModal;
  Escape[s_] := StringReplace[s, " " -> "^ "],
(* else *)
  $EditorModal = "${VISUAL:-xterm -e nano} ``";
  $Editor := $EditorModal <> " &";
  Escape[s_] := "\"" <> s <> "\"" ]

toform = "!" <> Escape[ToFileName[$FormRunBin, "ToForm"]] <> " > "


ParseOpt[f_, opt___] := (
  Message[f::optx, #, f]&/@ Complement[First/@ {opt}, #];
  # //. Level[{{opt}, Options[f]}, {2}]
)&[ First/@ Options[f] ]


tempnum = 1000 $KernelID + 1 /. HoldPattern[$KernelID] -> 0

FormFile[stub_, n_] := ToFileName[Directory[], stub <> ToString[n] <> ".frm"]


FormSeq[x_List] := StringTake[ToString[x, InputForm], {2, -2}]


root$[d_, x_] := x^(1/d)

FormPower[Rational[x_,y_], Rational[n_,d_]] := "root_"[d, (x y^(d-1))^n]/y^n

FormPower[x_?NumberQ, Rational[n_?Negative,d_]] := "root_"[d, (x^(d-1))^-n] x^n

FormPower[x_?NumberQ, Rational[n_, d_]] := "root_"[d, x^n]

FormPower[other__] := Power[other]


ToSym[x__] := ToExpression[StringJoin[ToString/@ Flatten[{x}]]]


FormWrite[hh_][lhs_ -> rhs_List] :=
Block[ {c = 0, LHS},
  LHS = ToSym[lhs, ++c]&/@ rhs;
  {lhs -> LHS, FormWrite[hh]/@ Thread[LHS -> rhs]}
]

FormWrite[hh_][lhs_ -> rhs_] := (
  Write[hh, "L ", lhs, " = ", rhs /. Power -> FormPower, ";"];
  WriteString[hh, "\n"];
  {}
)

f_FormWrite[rhs_] := f["expr" -> rhs]


Options[FormRun] = {
  FilePrefix -> "fr-",
  RetainFile -> False,
  EditCode -> False
}

FormRun[expr_, decl_String:"", cmd_String:"", opt___Rule] := 
Block[ {prefix, edit, retain, tempfile, hh, fdecl, lhs, res},
  {prefix, retain, edit} = ParseOpt[FormRun, opt];

  While[ FileType[tempfile = FormFile[prefix, tempnum]] =!= None,
    ++tempnum ];
  Print["\npreparing FORM code in " <> tempfile];
  hh = OpenWrite[toform <> Escape[tempfile],
    FormatType -> InputForm, PageWidth -> 73];

  fdecl = Union[Cases[expr,
    (h_Symbol)[___] :> h /; Context[h] =!= "System`", {-2}]];
  fdecl = If[Length[fdecl] === 0, {}, {"cf ", FormSeq[fdecl], ";\n"}];

  WriteString[hh, FormSetup <> fdecl <> "##\n" <> decl <> "\n##\n\n"];
  lhs = FormWrite[hh][expr];
  WriteString[hh, "##\n" <> cmd <> "\n\nprint;\n.end\n##\n"];

  Close[hh]

  Switch[ edit,
    True, Pause[1]; Run[StringForm[$Editor, tempfile]]; Pause[3],
    Modal, Pause[1]; Run[StringForm[$EditorModal, tempfile]] ];

  Print["running FORM... "];
  If[ Check[res = FormGet[$FormCmd, tempfile]; True, False] && !retain,
    DeleteFile[tempfile] ];
  Print["ok\n"];

  Flatten[{lhs, res}]
]

End[]

EndPackage[]

